@@ -26,6 +26,8 @@
 
 #include "bcachefs.h"
 
+#include "init/fs.h"
+
 #include "sb/io.h"
 #include "sb/members.h"
 
@@ -80,7 +82,8 @@ static void print_one_member(struct printbuf *out, sb_names sb_names,
 	printbuf_indent_sub(out, 2);
 }
 
-void bch2_sb_to_text_with_names(struct printbuf *out, struct bch_sb *sb,
+void bch2_sb_to_text_with_names(struct printbuf *out,
+				struct bch_fs *c, struct bch_sb *sb,
 				bool print_layout, unsigned fields, int field_only)
 {
 	CLASS(printbuf, uuid_buf)();
@@ -94,11 +97,11 @@ void bch2_sb_to_text_with_names(struct printbuf *out, struct bch_sb *sb,
 		struct bch_sb_field *f = bch2_sb_field_get_id(sb, field_only);
 
 		if (f)
-			__bch2_sb_field_to_text(out, sb, f);
+			__bch2_sb_field_to_text(out, c, sb, f);
 	} else {
 		printbuf_tabstop_push(out, 44);
 
-		bch2_sb_to_text(out, sb, print_layout,
+		bch2_sb_to_text(out, c, sb, print_layout,
 				fields & ~(BIT(BCH_SB_FIELD_members_v1)|
 					   BIT(BCH_SB_FIELD_members_v2)));
 
@@ -164,33 +167,40 @@ int cmd_show_super(int argc, char *argv[])
 	if (argc)
 		die("too many arguments");
 
+	CLASS(darray_const_str, devs)();
+	darray_push(&devs, dev);
+
 	struct bch_opts opts = bch2_opts_empty();
+	opt_set(opts, noexcl,		true);
+	opt_set(opts, nochanges,	true);
+	opt_set(opts, no_version_check,	true);
+	opt_set(opts, nostart,		true);
 
-	opt_set(opts, noexcl,	true);
-	opt_set(opts, nochanges, true);
-	opt_set(opts, no_version_check, true);
-
-	struct bch_sb_handle sb;
-	int ret = bch2_read_super(dev, &opts, &sb);
+	struct bch_fs *c = bch2_fs_open(&devs, &opts);
+	int ret = PTR_ERR_OR_ZERO(c);
 	if (ret)
 		die("Error opening %s: %s", dev, bch2_err_str(ret));
 
-	if (print_default_fields) {
-		fields |= bch2_sb_field_get(sb.sb, members_v2)
-			? BIT(BCH_SB_FIELD_members_v2)
-			: BIT(BCH_SB_FIELD_members_v1);
-		fields |= BIT(BCH_SB_FIELD_errors);
+	/* only one member will be online */
+	for_each_online_member(c, ca, 0) {
+		struct bch_sb *sb = ca->disk_sb.sb;
+
+		CLASS(printbuf, buf)();
+
+		buf.human_readable_units = true;
+
+		if (print_default_fields) {
+			fields |= bch2_sb_field_get(sb, members_v2)
+				? BIT(BCH_SB_FIELD_members_v2)
+				: BIT(BCH_SB_FIELD_members_v1);
+			fields |= BIT(BCH_SB_FIELD_errors);
+		}
+
+		bch2_sb_to_text_with_names(&buf, c, sb, print_layout, fields, field_only);
+		printf("%s", buf.buf);
 	}
 
-	struct printbuf buf = PRINTBUF;
-
-	buf.human_readable_units = true;
-
-	bch2_sb_to_text_with_names(&buf, sb.sb, print_layout, fields, field_only);
-	printf("%s", buf.buf);
-
-	bch2_free_super(&sb);
-	printbuf_exit(&buf);
+	bch2_fs_exit(c);
 	return 0;
 }
 
@@ -473,7 +483,7 @@ int cmd_recover_super(int argc, char *argv[])
 		: recover_super_from_member(args);
 
 	struct printbuf buf = PRINTBUF;
-	bch2_sb_to_text(&buf, sb, true, BIT_ULL(BCH_SB_FIELD_members_v2));
+	bch2_sb_to_text(&buf, NULL, sb, true, BIT_ULL(BCH_SB_FIELD_members_v2));
 
 	printf("Found superblock:\n%s\n", buf.buf);
 
